##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
#adapted from various trunk xmlrpc files jwashin 2005-06-09

__docformat__ = 'restructuredtext'

from minjson import read, write, ReadException, WriteException
from zExceptions.ExceptionFormatter import format_exception
from StringIO import StringIO
import sys, traceback, re
from zLOG import LOG, INFO, DEBUG
from cgi import FieldStorage

# this is used to identify incoming requests
request_content_type = 'application/json-rpc'

response_content_type = 'application/x-javascript'
response_encoding = None
# The below setting would be desired...
# but it breaks Opera 8.0.
## response_content_type = 'application/json-rpc'
## response_encoding = 'utf-8'

def get_response_content_type(encoding=response_encoding):
    result = response_content_type
    if encoding:
        result += ';charset=%s' % (encoding, )
    return result

def parse_input(data, encoding):
    """Parse input data and return a method path and argument tuple

    The data is a string.
    """
    # make the data to unicode
    if not isinstance(data, unicode):
        try:
            data = data.decode(encoding)
        except UnicodeDecodeError:
            data = data.decode('utf-8')
    data = read(data)
    functionstr = data[u'method'].encode('utf-8')
    if functionstr:
        method = functionstr.replace('.', '/')
    else:
        method = None
    jsonID = data[u'id']
    params = data[u'params']
    # Translate '.' to '/' in meth to represent object traversal.
    method = method.replace('.', '/')
    LOG('jsonserver', DEBUG, "processing request %s" % (data, ))
    return jsonID, method, params

class Response:
    '''Customized Response
    '''
    # use delegation, rather than inheritance

    def __init__(self, real, jsonID): 
        self.__dict__['_real'] = real
        self.__dict__['_jsonID'] = jsonID
        real._encode_unicode = self._encode_unicode

    def __getattr__(self, name): return getattr(self._real, name)
    def __setattr__(self, name, v): return setattr(self._real, name, v)
    def __delattr__(self, name): return delattr(self._real, name)

    def setBody(self, body, title='', is_error=0, bogus_str_search=None):
        """return
        {
        'id' : matches id in request
        'result' : the result or null if error
        'error' : the error or null if result
        }
        """
        if self._jsonID is None:
            self._real.setBody('')
            self._real.setStatus(204)
            LOG('jsonserver', DEBUG, "processing response 204 for id=%s" % (self._jsonID, ))
        else:
            body = premarshal(body)
            wrapper = {'id':self._jsonID}
            wrapper['result'] = body
            wrapper['error'] = None
            LOG('jsonserver', DEBUG, "processing response %s" % (wrapper, ))
            content_type = get_response_content_type()
            self._real.setHeader('content-type', content_type)
            cookedbody = write(wrapper)
            self._real.setBody(cookedbody)
            self._real.setStatus(200)
        return self

    def exception(self, fatal=0, info=None,
                  absuri_match=None, tag_search=None):
        # Fetch our exception info. t is type, v is value and tb is the
        # traceback object.
        if type(info) is type(()) and len(info)==3:
            t, v, tb = info
        else: 
            t, v, tb = sys.exc_info()

        LOG('jsonserver', ERROR, 'Exception caught:' + '\n'.join(format_exception(t, v, tb)))
        s = '%s: %s' % (getattr(t, '__name__', t), v)
        wrapper = {'id': self._jsonID}
        wrapper['result'] = None
        wrapper['error'] = s
        
        self._real.setHeader('content-type', response_content_type)
        self._real.setBody(write(wrapper))
        self._real.setStatus(200)

        #TODO What should really happen on error in a notification?
        
response=Response

def premarshal_dict(data):
    """return a non-proxied dict"""
    return dict([(premarshal(k), premarshal(v))
                 for (k, v) in data.items()])

def premarshal_list(data):
    """return a non-proxied list"""
    return map(premarshal, data)

#note: no dates or datetimes in json, but supported by xmlrpc
premarshal_dispatch_table = {
    dict: premarshal_dict,
    list: premarshal_list,
    tuple: premarshal_list,
    }

premarshal_dispatch = premarshal_dispatch_table.get

def premarshal(data):
    premarshaller = premarshal_dispatch(data.__class__)
    if premarshaller is not None:
        return premarshaller(data)
    return data

# --
# Patching processInputs of ZPublisher.HTTPRequest
# --

re_content_type= re.compile(r'charset\s*=\s*([^;]+)')

def processInputs(self, **kw):
    'Process request inputs'
    response=self.response
    environ=self.environ
    method=environ.get('REQUEST_METHOD','GET')

    if method != 'GET': fp=self.stdin
    else:               fp=None

    form=self.form
    other=self.other

    meth=None
    fs=FieldStorage(fp=fp,environ=environ,keep_blank_values=1)

    # Fix broken content types of Opera
    # this is because the xmlHTTPRequest method cannot set
    # the headers so it will always send text/xml;... ,
    # but we want to identify the requests regardless!
    if fs.headers.has_key('content-type') and \
            fs.headers['content-type'].startswith('text/xml') and \
            isinstance(fs.value, str) and fs.value.startswith('{"id":'):
        fs.headers['content-type'] = request_content_type
    if fs.headers.has_key('content-type') and \
            fs.headers['content-type'] == request_content_type \
            and method == 'POST':
        # get the content charset, suppose utf if not given.
        match = re_content_type.search(fs.headers['content-type'].lower())
        if match is not None:
            charset = match.group(1)
        else:
            charset = 'utf-8'
        #LOG('jsonserver', DEBUG, 'Considering request charset %s' % (repr(charset), None))
        jsonID, meth, self.args = parse_input(fs.value, charset)
        response = Response(response, jsonID)
        other['RESPONSE'] = self.response = response
        self.maybe_webdav_client = 0
        # continue with what used to be at the end of processInputs
        if environ.has_key('PATH_INFO'):
            path=environ['PATH_INFO']
            while path[-1:]=='/': path=path[:-1]
        else: path=''
        other['PATH_INFO']=path="%s/%s" % (path,meth)
        self._hacked_path=1
        return self
    else:
        self.stdin.seek(0)
        return self._processInputs_jsonrc_patched(**kw)

from ZPublisher.HTTPRequest import HTTPRequest
def patch_HTTPRequest():
    'This will patch HTTPRequest to enable json-rpc handling'
    HTTPRequest._processInputs_jsonrc_patched, HTTPRequest.processInputs = \
        HTTPRequest.processInputs, processInputs
    LOG('jsonserver', INFO, '*** Patching ZPublisher.HTTPRequest for json-rpc ***')
